@@ -1,36 +1,58 @@
-# SPDX-License-Identifier: BSD-3-Clause
-# Copyright (c) 2020 Deutsches Elektronen-Synchrotron DESY.
-# See LICENSE.txt for license details.
+###########################################################################
+#      ____  _____________  __    __  __ _           _____ ___   _        #
+#     / __ \/ ____/ ___/\ \/ /   |  \/  (_)__ _ _ __|_   _/ __| /_\  (R)  #
+#    / / / / __/  \__ \  \  /    | |\/| | / _| '_/ _ \| || (__ / _ \      #
+#   / /_/ / /___ ___/ /  / /     |_|  |_|_\__|_| \___/|_| \___/_/ \_\     #
+#  /_____/_____//____/  /_/      T  E  C  H  N  O  L  O  G  Y   L A B     #
+#                                                                         #
+#          Copyright 2021 Deutsches Elektronen-Synchrotron DESY.          #
+#                  SPDX-License-Identifier: BSD-3-Clause                  #
+#                                                                         #
+###########################################################################
 #
 # Terminal for the custom "serial over IPMB" protocol used by DESY MMC
-# 
+#
 # Based on pyserial miniterm, https://github.com/pyserial/pyserial/blob/master/serial/tools/miniterm.py
 # (C)2002-2020 Chris Liechti <cliechti@gmx.net>
+#
 
 import os
 import sys
 import threading
 import pyipmi
 import pyipmi.interfaces
+import pyipmi.errors
 import argparse
 import logging
 import time
 from enum import Enum
 from .__init__ import __version__
 
+# 32 byte max IPMB frame length minus headers etc.
+IPMB_MAX_PAYLOAD_LEN = 25
+IPMB_HEADER_LEN = 7
+
+
 class IpmiCode(Enum):
     SOI_CHANNEL_INFO = 0xf0
     SOI_SESSION_CTRL = 0xf1
     SOI_POLL_XCHG = 0xf2
 
+
 class IpmiConn():
     def __init__(self, mmc_addr, mch_url, ipmitool_mode=False):
         if ipmitool_mode:
-            self.interface = pyipmi.interfaces.create_interface('ipmitool', interface_type='lan')
+            self.interface = pyipmi.interfaces.create_interface(
+                'ipmitool', interface_type='lan')
         else:
-            self.interface = pyipmi.interfaces.create_interface('rmcp', keep_alive_interval=0)
+            self.interface = pyipmi.interfaces.create_interface(
+                'rmcp', keep_alive_interval=0)
         self.conn = self.mtca_mch_bridge_amc(mch_url, mmc_addr)
 
+    def __del__(self):
+        if hasattr(self, 'conn') and self.conn.session and self.conn.session.activated:
+            self.conn.session.close()
+
     '''
         From https://github.com/kontron/python-ipmi/blob/master/pyipmi/__init__.py#L111
 
@@ -50,18 +72,24 @@ class IpmiConn():
                     `-------------------´             `--------´
         `------------´     `---´        `---------------´
     '''
+
     def mtca_mch_bridge_amc(self, mch_url: str, amc_mmc_addr: int):
         '''
         Create a "double bridge" IPMI connection to talk directly to a AMC
         '''
         mtca_amc_double_bridge = [(0x81, 0x20, 0),
-                                 (0x20, 0x82, 7),
-                                 (0x20, amc_mmc_addr, None)]
+                                  (0x20, 0x82, 7),
+                                  (0x20, amc_mmc_addr, None)]
 
         conn = pyipmi.create_connection(self.interface)
         conn.session.set_session_type_rmcp(mch_url)
         conn.session.set_auth_type_user('', '')
-        conn.session.establish()
+        conn.interface.set_timeout(0.25)
+        try:
+            conn.session.establish()
+        except Exception as e:
+            raise RuntimeError(f'Couldn\'t connect to MCH {mch_url}: {e}')
+
         conn.target = pyipmi.Target(
             ipmb_address=amc_mmc_addr,
             routing=mtca_amc_double_bridge
@@ -81,7 +109,7 @@ class IpmiConn():
 
         raw_reply = self.conn.raw_command(0, 0x30, data)
         return raw_reply[0], raw_reply[1:]
-    
+
     def channel_list(self):
         '''
         Retrieve list of available "serial over IPMB" channels
@@ -101,17 +129,21 @@ class IpmiConn():
             ch_idx += 1
         return channels
 
-    def session_ctrl(self, channel, enable):
+    def session_ctrl(self, channel, enable, max_pkt_size=None):
         '''
         Open / close "serial over IPMB" session
         '''
         channel = int.to_bytes(channel, 1, byteorder='big')
         enable = b'\x01' if enable else b'\x00'
-        status, _ = self.raw_cmd(IpmiCode.SOI_SESSION_CTRL, channel + enable)
+        max_pkt_b = int.to_bytes(
+            max_pkt_size, 1, byteorder='big') if max_pkt_size is not None else b''
+        status, _ = self.raw_cmd(
+            IpmiCode.SOI_SESSION_CTRL, channel + enable + max_pkt_b)
         if status != 0:
-            print(f'session_ctrl returned 0x{status:02x}')
+            err = pyipmi.errors.CompletionCodeError(status)
+            print(f'session_ctrl: 0x{status:02x} ({err.cc_desc})')
         return status == 0
-    
+
     def poll_xchg(self, tx_data):
         '''
         Poll / exchange "serial over IPMB" data
@@ -119,9 +151,12 @@ class IpmiConn():
         # Assuming tx_data is not longer than one max. TX packet (if that happens, we have to implement splitting)
         return self.raw_cmd(IpmiCode.SOI_POLL_XCHG, tx_data)
 
+
 '''
 Console code based on pyserial/miniterm
 '''
+
+
 class ConsoleBase(object):
     """OS abstraction for console (input/output codec, no echo)"""
 
@@ -203,21 +238,21 @@ if os.name == 'nt':  # noqa
             'R': '\x1b[2~',  # INSERT
             'S': '\x1b[3~',  # DELETE
             'I': '\x1b[5~',  # PGUP
-            'Q': '\x1b[6~',  # PGDN        
+            'Q': '\x1b[6~',  # PGDN
         }
-        
+
         def __init__(self):
             super(Console, self).__init__()
             self._saved_ocp = ctypes.windll.kernel32.GetConsoleOutputCP()
             self._saved_icp = ctypes.windll.kernel32.GetConsoleCP()
             ctypes.windll.kernel32.SetConsoleOutputCP(65001)
             ctypes.windll.kernel32.SetConsoleCP(65001)
-            # ANSI handling available through SetConsoleMode since Windows 10 v1511 
+            # ANSI handling available through SetConsoleMode since Windows 10 v1511
             # https://en.wikipedia.org/wiki/ANSI_escape_code#cite_note-win10th2-1
             if platform.release() == '10' and int(platform.version().split('.')[2]) > 10586:
                 ENABLE_VIRTUAL_TERMINAL_PROCESSING = 0x0004
                 import ctypes.wintypes as wintypes
-                if not hasattr(wintypes, 'LPDWORD'): # PY2
+                if not hasattr(wintypes, 'LPDWORD'):  # PY2
                     wintypes.LPDWORD = ctypes.POINTER(wintypes.DWORD)
                 SetConsoleMode = ctypes.windll.kernel32.SetConsoleMode
                 GetConsoleMode = ctypes.windll.kernel32.GetConsoleMode
@@ -225,11 +260,14 @@ if os.name == 'nt':  # noqa
                 mode = wintypes.DWORD()
                 GetConsoleMode(GetStdHandle(-11), ctypes.byref(mode))
                 if (mode.value & ENABLE_VIRTUAL_TERMINAL_PROCESSING) == 0:
-                    SetConsoleMode(GetStdHandle(-11), mode.value | ENABLE_VIRTUAL_TERMINAL_PROCESSING)
+                    SetConsoleMode(GetStdHandle(-11), mode.value |
+                                   ENABLE_VIRTUAL_TERMINAL_PROCESSING)
                     self._saved_cm = mode
-            self.output = Out(sys.stdout.fileno()) # codecs.getwriter('UTF-8')(Out(sys.stdout.fileno()), 'replace')
+            # codecs.getwriter('UTF-8')(Out(sys.stdout.fileno()), 'replace')
+            self.output = Out(sys.stdout.fileno())
             # the change of the code page is not propagated to Python, manually fix it
-            sys.stderr = Out(sys.stderr.fileno()) # codecs.getwriter('UTF-8')(Out(sys.stderr.fileno()), 'replace')
+            # codecs.getwriter('UTF-8')(Out(sys.stderr.fileno()), 'replace')
+            sys.stderr = Out(sys.stderr.fileno())
             sys.stdout = self.output
             self.output.encoding = 'UTF-8'  # needed for input
 
@@ -237,8 +275,9 @@ if os.name == 'nt':  # noqa
             ctypes.windll.kernel32.SetConsoleOutputCP(self._saved_ocp)
             ctypes.windll.kernel32.SetConsoleCP(self._saved_icp)
             try:
-                ctypes.windll.kernel32.SetConsoleMode(ctypes.windll.kernel32.GetStdHandle(-11), self._saved_cm)
-            except AttributeError: # in case no _saved_cm
+                ctypes.windll.kernel32.SetConsoleMode(
+                    ctypes.windll.kernel32.GetStdHandle(-11), self._saved_cm)
+            except AttributeError:  # in case no _saved_cm
                 pass
 
         def getkey(self):
@@ -300,8 +339,11 @@ else:
     raise NotImplementedError(
         'Sorry no implementation for your platform ({}) available.'.format(sys.platform))
 
+
 class MMCterm(object):
-    def __init__(self, ipmi_conn):
+    STATUS_ERROR_RETRIES = 10
+
+    def __init__(self, ipmi_conn, max_pkt_size, polling_interval):
         self.console = Console()
         self.ipmi = ipmi_conn
         self.exit_character = chr(0x18)  # ctrl-x
@@ -310,6 +352,12 @@ class MMCterm(object):
         self.receiver_thread = None
         self.console_queue = []
         self.queue_lock = threading.Lock()
+        max_pkt_size = max_pkt_size or 32
+        self.polling_interval = polling_interval or 10
+        self.polling_interval /= 1000
+        self.max_payload_len = max_pkt_size - IPMB_HEADER_LEN
+        self.max_payload_len = max(self.max_payload_len, 1)
+        self.retries = MMCterm.STATUS_ERROR_RETRIES
 
     def _start_reader(self):
         """Start reader thread"""
@@ -327,7 +375,8 @@ class MMCterm(object):
         """start worker threads"""
         self.alive = True
         self._start_reader()
-        self.transmitter_thread = threading.Thread(target=self.writer, name='tx')
+        self.transmitter_thread = threading.Thread(
+            target=self.writer, name='tx')
         self.transmitter_thread.daemon = True
         self.transmitter_thread.start()
         self.console.setup()
@@ -363,19 +412,32 @@ class MMCterm(object):
 
             if len(tx_data) == 0 and len(rx_data) == 0:
                 # Don't flood the MCH with polling, if there's probably no data to exchange
-                time.sleep(0.01)
+                time.sleep(self.polling_interval)
 
-            # write user input to MMC, fetch MMC output to print
-            try:
-                status, rx_data = self.ipmi.poll_xchg(bytearray(tx_data))
-            except Exception as e:
-                self.reader_abort(f'pyipmi exception: {e}')
-                return
+            while True:
+                tx_part, tx_data = tx_data[:IPMB_MAX_PAYLOAD_LEN], tx_data[IPMB_MAX_PAYLOAD_LEN:]
+                # write user input to MMC, fetch MMC output to print
+                try:
+                    status, rx_data = self.ipmi.poll_xchg(bytearray(tx_part))
+                except Exception as e:
+                    self.reader_abort(f'pyipmi exception: {e}')
+                    return
 
-            if len(rx_data):
-                self.console.write_bytes(rx_data)
-            if status != 0:
-                self.reader_abort(f'Status error: {status:02x}')
+                if len(rx_data):
+                    self.console.write_bytes(rx_data)
+                if status != 0:
+                    self.retries -= 1
+                    # NAT MCH sometimes sends 0xd3 ("Destination unavailable")
+                    # Ignore it and retry if it doesn't happen N times in a row
+                    if status != 0xd3 or not self.retries:
+                        self.reader_abort(
+                            f'Status error: {status:02x}, giving up')
+
+                else:
+                    self.retries = MMCterm.STATUS_ERROR_RETRIES
+
+                if len(tx_data) == 0:
+                    break
 
     def writer(self):
         '''
@@ -395,53 +457,69 @@ class MMCterm(object):
                     break
                 else:
                     with self.queue_lock:
-                        self.console_queue += c.replace('\n', '\r').encode('utf-8')
+                        self.console_queue += c.replace('\n',
+                                                        '\r').encode('utf-8')
         except:
             self.alive = False
             raise
 
 
 def main():
-# example: ./mmcterm.py 0x74 -m 192.168.1.252
+    # example: ./mmcterm.py 0x74 -m 192.168.1.252
     parser = argparse.ArgumentParser(
         description='DESY MMC Serial over IPMB console'
     )
     parser.add_argument('mch_addr',
                         type=str,
                         help='IP address or hostname of MCH'
-    )
+                        )
     parser.add_argument('mmc_addr',
-                        type=lambda x: int(x,0),
+                        type=lambda x: int(x, 0),
                         help='IPMB-L address of MMC'
-    )
+                        )
     parser.add_argument('-v', '--version',
                         action='version',
                         version='%(prog)s ' + __version__
-    )
+                        )
     parser.add_argument('-c', '--channel',
                         type=int,
                         default=0,
                         help='console channel (default 0)'
-    )
+                        )
+    parser.add_argument('-t', '--interval',
+                        type=int,
+                        default=10,
+                        help='polling interval in ms (default 10)'
+                        )
     parser.add_argument('-l', '--list',
                         action='store_true',
                         help='list available channels'
-    )
+                        )
     parser.add_argument('-d', '--debug',
                         action='store_true',
                         help='pyipmi debug mode'
-    )
+                        )
     parser.add_argument('-i', '--ipmitool',
                         action='store_true',
                         help='make pyipmi use ipmitool instead of native rmcp'
-    )
+                        )
+    parser.add_argument('-m', '--max-pkt-size',
+                        type=int,
+                        help='max IPMB packet size to use'
+                        ' (Higher numbers give better performance, but can break depending on MCH model)'
+                        )
     args = parser.parse_args()
 
     if args.debug:
         pyipmi.logger.set_log_level(logging.DEBUG)
         pyipmi.logger.add_log_handler(logging.StreamHandler())
-    
-    conn = IpmiConn(args.mmc_addr, args.mch_addr, ipmitool_mode=args.ipmitool)
+
+    try:
+        conn = IpmiConn(args.mmc_addr, args.mch_addr,
+                        ipmitool_mode=args.ipmitool)
+    except Exception as e:
+        print(e)
+        sys.exit(1)
 
     if args.list:
         lst = conn.channel_list()
@@ -453,11 +531,11 @@ def main():
             print(f'channel {l[0]}: {l[1]}')
         sys.exit(0)
 
-    if not conn.session_ctrl(args.channel, True):
+    if not conn.session_ctrl(args.channel, True, args.max_pkt_size):
         print(f'Could not open session for channel {args.channel}')
         sys.exit(-1)
 
-    mmcterm = MMCterm(conn)
+    mmcterm = MMCterm(conn, args.max_pkt_size, args.interval)
 
     print("Press Ctrl-x to exit")
     mmcterm.start()
@@ -469,5 +547,6 @@ def main():
 
     conn.session_ctrl(args.channel, False)
 
+
 if __name__ == '__main__':
     main()
